@@ -25,7 +25,7 @@
 
 /******************************************************************************/
 
-import { StaticFilteringParser } from '../static-filtering-parser.js';
+import * as sfp from '../static-filtering-parser.js';
 
 /******************************************************************************/
 
@@ -39,336 +39,219 @@ let hintHelperRegistered = false;
 /******************************************************************************/
 
 CodeMirror.defineMode('ubo-static-filtering', function() {
-    if ( StaticFilteringParser instanceof Object === false ) { return; }
-    const parser = new StaticFilteringParser({ interactive: true });
+    if ( sfp.AstFilterParser instanceof Object === false ) { return; }
+    const astParser = new sfp.AstFilterParser({
+        interactive: true,
+        nativeCssHas: vAPI.webextFlavor.env.includes('native_css_has'),
+    });
+    const astWalker = astParser.getWalker();
+    let currentWalkerNode = 0;
+    let lastNetOptionType = 0;
 
-    const reURL = /\bhttps?:\/\/\S+/;
-    const rePreparseDirectives = /^!#(?:if|endif|include )\b/;
-    const rePreparseIfDirective = /^(!#if ?)(.*)$/;
-    let parserSlot = 0;
-    let netOptionValueMode = false;
-
-    const colorCommentSpan = function(stream) {
-        const { string, pos } = stream;
-        if ( rePreparseDirectives.test(string) === false ) {
-            const match = reURL.exec(string.slice(pos));
-            if ( match !== null ) {
-                if ( match.index === 0 ) {
-                    stream.pos += match[0].length;
-                    return 'comment link';
-                }
-                stream.pos += match.index;
-                return 'comment';
-            }
-            stream.skipToEnd();
-            return 'comment';
-        }
-        const match = rePreparseIfDirective.exec(string);
-        if ( match === null ) {
-            stream.skipToEnd();
-            return 'variable strong';
-        }
-        if ( pos < match[1].length ) {
-            stream.pos += match[1].length;
-            return 'variable strong';
-        }
-        stream.skipToEnd();
-        if ( match[1].endsWith(' ') === false ) {
-            return 'error strong';
-        }
-        if ( preparseDirectiveTokens.size === 0 ) {
-            return 'positive strong';
-        }
-        let token = match[2];
-        const not = token.startsWith('!');
-        if ( not ) {
-            token = token.slice(1);
-        }
-        if ( preparseDirectiveTokens.has(token) === false ) {
-            return 'error strong';
-        }
-        if ( not !== preparseDirectiveTokens.get(token) ) {
-            return 'positive strong';
-        }
-        return 'negative strong';
+    const redirectTokenStyle = node => {
+        const rawToken = astParser.getNodeString(node);
+        const { token } = sfp.parseRedirectValue(rawToken);
+        return redirectNames.has(token) ? 'value' : 'value warning';
     };
 
-    const colorExtHTMLPatternSpan = function(stream) {
-        const { i } = parser.patternSpan;
-        if ( stream.pos === parser.slices[i+1] ) {
-            stream.pos += 1;
-            return 'def';
-        }
-        stream.skipToEnd();
-        return 'variable';
-    };
-
-    const colorExtScriptletPatternSpan = function(stream) {
-        const { pos, string } = stream;
-        const { i, len } = parser.patternSpan;
-        const patternBeg = parser.slices[i+1];
-        if ( pos === patternBeg ) {
-            stream.pos = pos + 4;
-            return 'def';
-        }
-        if ( len > 3 ) {
-            if ( pos === patternBeg + 4 ) {
-                const match = /^[^,)]+/.exec(string.slice(pos));
-                const token = match && match[0].trim();
-                if ( token && scriptletNames.has(token) === false ) {
-                    stream.pos = pos + match[0].length;
-                    return 'warning';
-                }
-            }
-            const r = parser.slices[i+len+1] - 1;
-            if ( pos < r ) {
-                stream.pos = r;
-                return 'variable';
-            }
-            if ( pos === r ) {
-                stream.pos = pos + 1;
-                return 'def';
-            }
-        }
-        stream.skipToEnd();
-        return 'variable';
-    };
-
-    const colorExtPatternSpan = function(stream) {
-        if ( (parser.flavorBits & parser.BITFlavorExtScriptlet) !== 0 ) {
-            return colorExtScriptletPatternSpan(stream);
-        }
-        if ( (parser.flavorBits & parser.BITFlavorExtHTML) !== 0 ) {
-            return colorExtHTMLPatternSpan(stream);
-        }
-        stream.skipToEnd();
-        return 'variable';
-    };
-
-    const colorExtSpan = function(stream) {
-        if ( parserSlot < parser.optionsAnchorSpan.i ) {
-            const style = (parser.slices[parserSlot] & parser.BITComma) === 0
-                ? 'value'
-                : 'def';
-            stream.pos += parser.slices[parserSlot+2];
-            parserSlot += 3;
-            return style;
-        }
-        if (
-            parserSlot >= parser.optionsAnchorSpan.i &&
-            parserSlot < parser.patternSpan.i
-        ) {
-            const style = (parser.flavorBits & parser.BITFlavorException) !== 0
-                ? 'tag'
-                : 'def';
-            stream.pos += parser.slices[parserSlot+2];
-            parserSlot += 3;
-            return `${style} strong`;
-        }
-        if (
-            parserSlot >= parser.patternSpan.i &&
-            parserSlot < parser.rightSpaceSpan.i
-        ) {
-            return colorExtPatternSpan(stream);
-        }
-        stream.skipToEnd();
-        return null;
-    };
-
-    const colorNetOptionValueSpan = function(stream, bits) {
-        const { pos, string } = stream;
-        let style;
-        // Warn about unknown redirect tokens.
-        if (
-            string.charCodeAt(pos - 1) === 0x3D /* '=' */ &&
-            /[$,](redirect(-rule)?|rewrite)=$/.test(string.slice(0, pos))
-        ) {
-            style = 'value';
-            const end = parser.skipUntil(
-                parserSlot,
-                parser.commentSpan.i,
-                parser.BITComma
-            );
-            const raw = parser.strFromSlices(parserSlot, end - 3);
-            const { token } = StaticFilteringParser.parseRedirectValue(raw);
-            if ( redirectNames.has(token) === false ) {
-                style += ' warning';
-            }
-            stream.pos += raw.length;
-            parserSlot = end;
-            return style;
-        }
-        if ( (bits & parser.BITTilde) !== 0 ) {
-            style = 'keyword strong';
-        } else if ( (bits & parser.BITPipe) !== 0 ) {
-            style = 'def';
-        }
-        stream.pos += parser.slices[parserSlot+2];
-        parserSlot += 3;
-        return style || 'value';
-    };
-
-    // https://github.com/uBlockOrigin/uBlock-issues/issues/760#issuecomment-951146371
-    //   Quick fix: auto-escape commas.
-    const colorNetOptionSpan = function(stream) {
-        const [ slotBits, slotPos, slotLen ] =
-            parser.slices.slice(parserSlot, parserSlot+3);
-        if ( (slotBits & parser.BITComma) !== 0 ) {
-            if ( /^,\d*?\}/.test(parser.raw.slice(slotPos)) === false ) {
-                netOptionValueMode = false;
-                stream.pos += slotLen;
-                parserSlot += 3;
-                return 'def strong';
-            }
-        }
-        if ( netOptionValueMode ) {
-            return colorNetOptionValueSpan(stream, slotBits);
-        }
-        if ( (slotBits & parser.BITTilde) !== 0 ) {
-            stream.pos += slotLen;
-            parserSlot += 3;
-            return 'keyword strong';
-        }
-        if ( (slotBits & parser.BITEqual) !== 0 ) {
-            netOptionValueMode = true;
-            stream.pos += slotLen;
-            parserSlot += 3;
-            return 'def';
-        }
-        parserSlot = parser.skipUntil(
-            parserSlot,
-            parser.commentSpan.i,
-            parser.BITComma | parser.BITEqual
-        );
-        stream.pos = parser.slices[parserSlot+1];
-        return 'def';
-    };
-
-    const colorNetSpan = function(stream) {
-        if ( parserSlot < parser.exceptionSpan.i ) {
-            stream.pos += parser.slices[parserSlot+2];
-            parserSlot += 3;
-            return null;
-        }
-        if (
-            parserSlot === parser.exceptionSpan.i &&
-            parser.exceptionSpan.len !== 0
-        ) {
-            stream.pos += parser.slices[parserSlot+2];
-            parserSlot += 3;
-            return 'tag strong';
-        }
-        if (
-            parserSlot === parser.patternLeftAnchorSpan.i &&
-            parser.patternLeftAnchorSpan.len !== 0 ||
-            parserSlot === parser.patternRightAnchorSpan.i &&
-            parser.patternRightAnchorSpan.len !== 0
-        ) {
-            stream.pos += parser.slices[parserSlot+2];
-            parserSlot += 3;
-            return 'keyword strong';
-        }
-        if (
-            parserSlot >= parser.patternSpan.i &&
-            parserSlot < parser.optionsAnchorSpan.i
-        ) {
-            if ( parser.patternIsRegex() ) {
-                stream.pos = parser.slices[parser.optionsAnchorSpan.i+1];
-                parserSlot = parser.optionsAnchorSpan.i;
-                return parser.patternIsTokenizable()
-                    ? 'variable notice'
-                    : 'variable warning';
-            }
-            if ( (parser.slices[parserSlot] & (parser.BITAsterisk | parser.BITCaret)) !== 0 ) {
-                stream.pos += parser.slices[parserSlot+2];
-                parserSlot += 3;
-                return 'keyword strong';
-            }
-            const nextSlot = parser.skipUntil(
-                parserSlot + 3,
-                parser.patternRightAnchorSpan.i,
-                parser.BITAsterisk | parser.BITCaret
-            );
-            stream.pos = parser.slices[nextSlot+1];
-            parserSlot = nextSlot;
-            return 'variable';
-        }
-        if (
-            parserSlot === parser.optionsAnchorSpan.i &&
-            parserSlot < parser.optionsSpan.i !== 0
-        ) {
-            stream.pos += parser.slices[parserSlot+2];
-            parserSlot += 3;
-            return 'def strong';
-        }
-        if (
-            parserSlot >= parser.optionsSpan.i &&
-            parserSlot < parser.commentSpan.i
-        ) {
-            return colorNetOptionSpan(stream);
-        }
-        if (
-            parserSlot >= parser.commentSpan.i &&
-            parser.commentSpan.len !== 0
-        ) {
-            stream.skipToEnd();
-            return 'comment';
-        }
-        stream.skipToEnd();
-        return null;
-    };
-
-    const colorSpan = function(stream) {
-        if ( parser.category === parser.CATNone || parser.shouldIgnore() ) {
-            stream.skipToEnd();
-            return 'comment';
-        }
-        if ( parser.category === parser.CATComment ) {
-            return colorCommentSpan(stream);
-        }
-        if ( (parser.slices[parserSlot] & parser.BITError) !== 0 ) {
-            stream.pos += parser.slices[parserSlot+2];
-            parserSlot += 3;
+    const colorFromAstNode = function() {
+        if ( astParser.nodeIsEmptyString(currentWalkerNode) ) { return '+'; }
+        if ( astParser.getNodeFlags(currentWalkerNode, sfp.NODE_FLAG_ERROR) !== 0 ) {
             return 'error';
         }
-        if ( (parser.slices[parserSlot] & parser.BITIgnore) !== 0 ) {
-            stream.pos += parser.slices[parserSlot+2];
-            parserSlot += 3;
-            return 'comment';
-        }
-        if ( parser.category === parser.CATStaticExtFilter ) {
-            const style = colorExtSpan(stream) || '';
-            let flavor = '';
-            if ( (parser.flavorBits & parser.BITFlavorExtCosmetic) !== 0 ) {
-                flavor = 'line-cm-ext-dom';
-            } else if ( (parser.flavorBits & parser.BITFlavorExtScriptlet) !== 0 ) {
-                flavor = 'line-cm-ext-js';
-            } else if ( (parser.flavorBits & parser.BITFlavorExtHTML) !== 0 ) {
-                flavor = 'line-cm-ext-html';
+        const nodeType = astParser.getNodeType(currentWalkerNode);
+        switch ( nodeType ) {
+            case sfp.NODE_TYPE_WHITESPACE:
+                return '';
+            case sfp.NODE_TYPE_COMMENT:
+                if ( astWalker.canGoDown() ) { break; }
+                return 'comment';
+            case sfp.NODE_TYPE_COMMENT_URL:
+                return 'comment link';
+            case sfp.NODE_TYPE_IGNORE:
+                return 'comment';
+            case sfp.NODE_TYPE_PREPARSE_DIRECTIVE:
+            case sfp.NODE_TYPE_PREPARSE_DIRECTIVE_VALUE:
+                return 'directive';
+            case sfp.NODE_TYPE_PREPARSE_DIRECTIVE_IF_VALUE: {
+                if ( preparseDirectiveTokens.size === 0 ) {
+                    return 'positive strong';
+                }
+                const raw = astParser.getNodeString(currentWalkerNode);
+                const not = raw.startsWith('!');
+                const token = not ? raw.slice(1) : raw;
+                if ( preparseDirectiveTokens.has(token) === false ) {
+                    return 'error strong';
+                }
+                return not === preparseDirectiveTokens.get(token)
+                    ? 'negative strong'
+                    : 'positive strong';
             }
-            return `${flavor} ${style}`.trim();
+            case sfp.NODE_TYPE_EXT_OPTIONS_ANCHOR:
+                return astParser.getFlags(sfp.AST_FLAG_IS_EXCEPTION)
+                    ? 'tag strong'
+                    : 'def strong';
+            case sfp.NODE_TYPE_EXT_DECORATION:
+                return 'def';
+            case sfp.NODE_TYPE_EXT_PATTERN_RAW:
+                if ( astWalker.canGoDown() ) { break; }
+                return 'variable';
+            case sfp.NODE_TYPE_EXT_PATTERN_COSMETIC:
+            case sfp.NODE_TYPE_EXT_PATTERN_HTML:
+                return 'variable';
+            case sfp.NODE_TYPE_EXT_PATTERN_RESPONSEHEADER:
+            case sfp.NODE_TYPE_EXT_PATTERN_SCRIPTLET:
+                if ( astWalker.canGoDown() ) { break; }
+                return 'variable';
+            case sfp.NODE_TYPE_EXT_PATTERN_SCRIPTLET_TOKEN: {
+                const token = astParser.getNodeString(currentWalkerNode);
+                if ( scriptletNames.has(token) === false ) {
+                    return 'warning';
+                }
+                return 'variable';
+            }
+            case sfp.NODE_TYPE_EXT_PATTERN_SCRIPTLET_ARG:
+                return 'variable';
+            case sfp.NODE_TYPE_NET_EXCEPTION:
+                return 'tag strong';
+            case sfp.NODE_TYPE_NET_PATTERN:
+                if ( astWalker.canGoDown() ) { break; }
+                if ( astParser.isRegexPattern() ) {
+                    if ( astParser.getNodeFlags(currentWalkerNode, sfp.NODE_FLAG_PATTERN_UNTOKENIZABLE) !== 0 ) {
+                        return 'variable warning';
+                    }
+                    return 'variable notice';
+                }
+                return 'variable';
+            case sfp.NODE_TYPE_NET_PATTERN_PART:
+                return 'variable';
+            case sfp.NODE_TYPE_NET_PATTERN_PART_SPECIAL:
+                return 'keyword strong';
+            case sfp.NODE_TYPE_NET_PATTERN_PART_UNICODE:
+                return 'variable unicode';
+            case sfp.NODE_TYPE_NET_PATTERN_LEFT_HNANCHOR:
+            case sfp.NODE_TYPE_NET_PATTERN_LEFT_ANCHOR:
+            case sfp.NODE_TYPE_NET_PATTERN_RIGHT_ANCHOR:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_NOT:
+                return 'keyword strong';
+            case sfp.NODE_TYPE_NET_OPTIONS_ANCHOR:
+            case sfp.NODE_TYPE_NET_OPTION_SEPARATOR:
+                lastNetOptionType = 0;
+                return 'def strong';
+            case sfp.NODE_TYPE_NET_OPTION_NAME_UNKNOWN:
+                lastNetOptionType = 0;
+                return 'error';
+            case sfp.NODE_TYPE_NET_OPTION_NAME_1P:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_STRICT1P:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_3P:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_STRICT3P:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_ALL:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_BADFILTER:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_CNAME:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_CSP:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_CSS:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_DENYALLOW:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_DOC:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_EHIDE:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_EMPTY:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_FONT:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_FRAME:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_FROM:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_GENERICBLOCK:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_GHIDE:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_HEADER:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_IMAGE:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_IMPORTANT:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_INLINEFONT:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_INLINESCRIPT:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_MATCHCASE:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_MEDIA:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_METHOD:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_MP4:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_NOOP:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_OBJECT:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_OTHER:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_PING:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_POPUNDER:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_POPUP:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_REDIRECT:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_REDIRECTRULE:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_REMOVEPARAM:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_SCRIPT:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_SHIDE:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_TO:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_XHR:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_WEBRTC:
+            case sfp.NODE_TYPE_NET_OPTION_NAME_WEBSOCKET:
+                lastNetOptionType = nodeType;
+                return 'def';
+            case sfp.NODE_TYPE_NET_OPTION_ASSIGN:
+                return 'def';
+            case sfp.NODE_TYPE_NET_OPTION_VALUE:
+                if ( astWalker.canGoDown() ) { break; }
+                switch ( lastNetOptionType ) {
+                    case sfp.NODE_TYPE_NET_OPTION_NAME_REDIRECT:
+                    case sfp.NODE_TYPE_NET_OPTION_NAME_REDIRECTRULE:
+                        return redirectTokenStyle(currentWalkerNode);
+                    default:
+                        break;
+                }
+                return 'value';
+            case sfp.NODE_TYPE_OPTION_VALUE_NOT:
+                return 'keyword strong';
+            case sfp.NODE_TYPE_OPTION_VALUE_DOMAIN:
+                return 'value';
+            case sfp.NODE_TYPE_OPTION_VALUE_SEPARATOR:
+                return 'def';
+            default:
+                break;
         }
-        if ( parser.category === parser.CATStaticNetFilter ) {
-            const style = colorNetSpan(stream);
-            return style ? `line-cm-net ${style}` : 'line-cm-net';
-        }
-        stream.skipToEnd();
-        return null;
+        return '+';
     };
 
     return {
         lineComment: '!',
         token: function(stream) {
-            let style = '';
             if ( stream.sol() ) {
-                parser.analyze(stream.string);
-                parser.analyzeExtra();
-                parserSlot = 0;
-                netOptionValueMode = false;
+                astParser.parse(stream.string);
+                if ( astParser.getFlags(sfp.AST_FLAG_UNSUPPORTED) !== 0 ) {
+                    stream.skipToEnd();
+                    return 'error';
+                }
+                if ( astParser.getType() === sfp.AST_TYPE_NONE ) {
+                    stream.skipToEnd();
+                    return 'comment';
+                }
+                currentWalkerNode = astWalker.reset();
+            } else {
+                currentWalkerNode = astWalker.next();
             }
-            style += colorSpan(stream) || '';
-            if ( (parser.flavorBits & parser.BITFlavorError) !== 0 ) {
-                style += ' line-background-error';
+            let style = '';
+            while ( currentWalkerNode !== 0 ) {
+                style = colorFromAstNode(stream);
+                if ( style !== '+' ) { break; }
+                currentWalkerNode = astWalker.next();
+            }
+            if ( style === '+' ) {
+                stream.skipToEnd();
+                return null;
+            }
+            stream.pos = astParser.getNodeStringEnd(currentWalkerNode);
+            if ( astParser.isNetworkFilter() ) {
+                return style ? `line-cm-net ${style}` : 'line-cm-net';
+            }
+            if ( astParser.isExtendedFilter() ) {
+                let flavor = '';
+                if ( astParser.isCosmeticFilter() ) {
+                    flavor = 'line-cm-ext-dom';
+                } else if ( astParser.isScriptletFilter() ) {
+                    flavor = 'line-cm-ext-js';
+                } else if ( astParser.isHtmlFilter() ) {
+                    flavor = 'line-cm-ext-html';
+                }
+                if ( flavor !== '' ) {
+                    style = `${flavor} ${style}`;
+                }
             }
             style = style.trim();
             return style !== '' ? style : null;
@@ -406,9 +289,7 @@ CodeMirror.defineMode('ubo-static-filtering', function() {
                 initHints();
             }
         },
-        get parser() {
-            return parser;
-        },
+        parser: astParser,
     };
 });
 
@@ -418,11 +299,14 @@ CodeMirror.defineMode('ubo-static-filtering', function() {
 //   https://codemirror.net/demo/complete.html
 
 const initHints = function() {
-    if ( StaticFilteringParser instanceof Object === false ) { return; }
+    if ( sfp.AstFilterParser instanceof Object === false ) { return; }
 
-    const parser = new StaticFilteringParser();
+    const astParser = new sfp.AstFilterParser({
+        interactive: true,
+        nativeCssHas: vAPI.webextFlavor.env.includes('native_css_has'),
+    });
     const proceduralOperatorNames = new Map(
-        Array.from(parser.proceduralOperatorTokens)
+        Array.from(sfp.proceduralOperatorTokens)
              .filter(item => (item[1] & 0b01) !== 0)
     );
     const excludedHints = new Set([
@@ -443,23 +327,89 @@ const initHints = function() {
             if ( text.startsWith(seed) === false ) { continue; }
             out.push(hint);
         }
-        // If no match, try again with a different heuristic
-        if ( out.length === 0 ) {
-            for ( const hint of hints ) {
-                const text = hint instanceof Object
-                    ? hint.displayText || hint.text
-                    : hint;
-                if ( seedLeft.length === 1 ) {
-                    if ( text.startsWith(seedLeft) === false ) { continue; }
-                } else if ( text.includes(seed) === false ) { continue; }
-                out.push(hint);
-            }
+        if ( out.length !== 0 ) {
+            return {
+                from: { line: cursor.line, ch: cursor.ch - seedLeft.length },
+                to: { line: cursor.line, ch: cursor.ch + seedRight.length },
+                list: out,
+            };
+        }
+        // If no match, try again with a different heuristic: valid hints are
+        // those matching left seed, not matching right seed but right seed is
+        // found to be a valid hint. This is to take care of cases like:
+        //
+        //     *$script,redomain=example.org
+        //                ^
+        //                + cursor position
+        //
+        // In such case, [ redirect=, redirect-rule= ] should be returned
+        // as valid hints.
+        for ( const hint of hints ) {
+            const text = hint instanceof Object
+                ? hint.displayText || hint.text
+                : hint;
+            if ( seedLeft.length === 0 ) { continue; }
+            if ( text.startsWith(seedLeft) === false ) { continue; }
+            if ( hints.includes(seedRight) === false ) { continue; }
+            out.push(hint);
+        }
+        if ( out.length !== 0 ) {
+            return {
+                from: { line: cursor.line, ch: cursor.ch - seedLeft.length },
+                to: { line: cursor.line, ch: cursor.ch },
+                list: out,
+            };
+        }
+        // If no match, try again with a different heuristic: valid hints are
+        // those containing seed as a substring. This is to take care of cases
+        // like:
+        //
+        //     *$script,redirect=gif
+        //                       ^
+        //                       + cursor position
+        //
+        // In such case, [ 1x1.gif, 1x1-transparent.gif ] should be returned
+        // as valid hints.
+        for ( const hint of hints ) {
+            const text = hint instanceof Object
+                ? hint.displayText || hint.text
+                : hint;
+            if ( seedLeft.length === 1 ) {
+                if ( text.startsWith(seedLeft) === false ) { continue; }
+            } else if ( text.includes(seed) === false ) { continue; }
+            out.push(hint);
+        }
+        if ( out.length !== 0 ) {
+            return {
+                from: { line: cursor.line, ch: cursor.ch - seedLeft.length },
+                to: { line: cursor.line, ch: cursor.ch + seedRight.length },
+                list: out,
+            };
+        }
+        // If still no match, try again with a different heuristic: valid hints
+        // are those containing left seed as a substring. This is to take care
+        // of cases like:
+        //
+        //     *$script,redirect=gifdomain=example.org
+        //                          ^
+        //                          + cursor position
+        //
+        // In such case, [ 1x1.gif, 1x1-transparent.gif ] should be returned
+        // as valid hints.
+        for ( const hint of hints ) {
+            const text = hint instanceof Object
+                ? hint.displayText || hint.text
+                : hint;
+            if ( text.includes(seedLeft) === false ) { continue; }
+            out.push(hint);
+        }
+        if ( out.length !== 0 ) {
+            return {
+                from: { line: cursor.line, ch: cursor.ch - seedLeft.length },
+                to: { line: cursor.line, ch: cursor.ch },
+                list: out,
+            };
         }
-        return {
-            from: { line: cursor.line, ch: cursor.ch - seedLeft.length },
-            to: { line: cursor.line, ch: cursor.ch + seedRight.length },
-            list: out,
-        };
     };
 
     const getOriginHints = function(cursor, line, suffix = '') {
@@ -491,16 +441,16 @@ const initHints = function() {
         }
         const assignPos = seedRight.indexOf('=');
         if ( assignPos !== -1 ) { seedRight = seedRight.slice(0, assignPos); }
-        const isException = parser.isException();
+        const isException = astParser.isException();
         const hints = [];
-        for ( let [ text, bits ] of parser.netOptionTokenDescriptors ) {
+        for ( let [ text, desc ] of sfp.netOptionTokenDescriptors ) {
             if ( excludedHints.has(text) ) { continue; }
-            if ( isNegated && (bits & parser.OPTCanNegate) === 0 ) { continue; }
+            if ( isNegated && desc.canNegate !== true ) { continue; }
             if ( isException ) {
-                if ( (bits & parser.OPTBlockOnly) !== 0 ) { continue; }
+                if ( desc.blockOnly ) { continue; }
             } else {
-                if ( (bits & parser.OPTAllowOnly) !== 0 ) { continue; }
-                if ( (assignPos === -1) && (bits & parser.OPTMustAssign) !== 0 ) {
+                if ( desc.allowOnly ) { continue; }
+                if ( (assignPos === -1) && desc.mustAssign ) {
                     text += '=';
                 }
             }
@@ -519,8 +469,11 @@ const initHints = function() {
     };
 
     const getNetHints = function(cursor, line) {
+        const patternNode = astParser.getBranchFromType(sfp.NODE_TYPE_NET_PATTERN_RAW);
+        if ( patternNode === 0 ) { return; }
+        const patternEnd = astParser.getNodeStringEnd(patternNode);
         const beg = cursor.ch;
-        if ( beg <= parser.slices[parser.optionsAnchorSpan.i+1] ) {
+        if ( beg <= patternEnd ) {
             return getNetPatternHints(cursor, line);
         }
         const lineBefore = line.slice(0, beg);
@@ -581,7 +534,7 @@ const initHints = function() {
         const matchRight = /^([^)]*)/.exec(line.slice(beg));
         if ( matchLeft === null || matchRight === null ) { return; }
         const hints = [];
-        for ( const hint of parser.removableHTTPHeaders ) {
+        for ( const hint of sfp.removableHTTPHeaders ) {
             hints.push(hint);
         }
         return pickBestHints(cursor, matchLeft[1], matchRight[1], hints);
@@ -628,29 +581,29 @@ const initHints = function() {
     CodeMirror.registerHelper('hint', 'ubo-static-filtering', function(cm) {
         const cursor = cm.getCursor();
         const line = cm.getLine(cursor.line);
-        parser.analyze(line);
-        if ( parser.category === parser.CATStaticExtFilter ) {
+        astParser.parse(line);
+        if ( astParser.isExtendedFilter() ) {
+            const anchorNode = astParser.getBranchFromType(sfp.NODE_TYPE_EXT_OPTIONS_ANCHOR);
+            if ( anchorNode === 0 ) { return; }
             let hints;
-            if ( cursor.ch <= parser.slices[parser.optionsAnchorSpan.i+1] ) {
+            if ( cursor.ch <= astParser.getNodeStringBeg(anchorNode) ) {
                 hints = getOriginHints(cursor, line);
-            } else if ( parser.hasFlavor(parser.BITFlavorExtScriptlet) ) {
+            } else if ( astParser.isScriptletFilter() ) {
                 hints = getExtScriptletHints(cursor, line);
-            } else if ( parser.hasFlavor(parser.BITFlavorExtResponseHeader) ) {
+            } else if ( astParser.isResponseheaderFilter() ) {
                 hints = getExtHeaderHints(cursor, line);
             } else {
                 hints = getExtSelectorHints(cursor, line);
             }
             return hints;
         }
-        if ( parser.category === parser.CATStaticNetFilter ) {
+        if ( astParser.isNetworkFilter() ) {
             return getNetHints(cursor, line);
         }
-        if ( parser.category === parser.CATComment ) {
+        if ( astParser.isComment() ) {
             return getCommentHints(cursor, line);
         }
-        if ( parser.category === parser.CATNone ) {
-            return getOriginHints(cursor, line);
-        }
+        return getOriginHints(cursor, line);
     });
 };
 
