@@ -23,12 +23,10 @@
 
 'use strict';
 
-/******************************************************************************/
-
 import './codemirror/ubo-static-filtering.js';
 
 import { hostnameFromURI } from './uri-utils.js';
-import { StaticFilteringParser } from './static-filtering-parser.js';
+import * as sfp from './static-filtering-parser.js';
 
 /******************************************************************************/
 /******************************************************************************/
@@ -112,14 +110,12 @@ const rawFilterFromTextarea = function() {
 const filterFromTextarea = function() {
     const filter = rawFilterFromTextarea();
     if ( filter === '' ) { return ''; }
-    const sfp = staticFilteringParser;
-    sfp.analyze(filter);
-    sfp.analyzeExtra();
-    if (
-        sfp.category !== sfp.CATStaticExtFilter &&
-        sfp.category !== sfp.CATStaticNetFilter ||
-        sfp.shouldDiscard()
-    ) {
+    const parser = staticFilteringParser;
+    parser.parse(filter);
+    if ( parser.isFilter() === false ) { return '!'; }
+    if ( parser.isExtendedFilter() ) {
+        if ( parser.isCosmeticFilter() === false ) { return '!'; }
+    } else if ( parser.isNetworkFilter() === false ) {
         return '!';
     }
     return filter;
@@ -604,6 +600,11 @@ const onStartMoving = (( ) => {
     let rMax = 0, bMax = 0;
     let timer;
 
+    const eatEvent = function(ev) {
+        ev.stopPropagation();
+        ev.preventDefault();
+    };
+
     const move = ( ) => {
         timer = undefined;
         const r1 = Math.min(Math.max(r0 - mx1 + mx0, 2), rMax);
@@ -708,13 +709,6 @@ const svgListening = (( ) => {
 
 /******************************************************************************/
 
-const eatEvent = function(ev) {
-    ev.stopPropagation();
-    ev.preventDefault();
-};
-
-/******************************************************************************/
-
 // Create lists of candidate filters. This takes into account whether the
 // current mode is narrow or broad.
 
@@ -834,7 +828,7 @@ const startPicker = function() {
     $id('candidateFilters').addEventListener('click', onCandidateClicked);
     $stor('#resultsetDepth input').addEventListener('input', onDepthChanged);
     $stor('#resultsetSpecificity input').addEventListener('input', onSpecificityChanged);
-    staticFilteringParser = new StaticFilteringParser({
+    staticFilteringParser = new sfp.AstFilterParser({
         interactive: true,
         nativeCssHas: vAPI.webextFlavor.env.includes('native_css_has'),
     });
